/*
 * Linux USB support
 *
 * Written by:
 * Copyright 2000, Johannes Erdfelt <johannes@erdfelt.com>
 *
 * This library is covered by the LGPL.
 */

#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h> 
#include <fcntl.h> 
#include <errno.h> 
#include <sys/time.h>
#include <dirent.h>

#include "linux.h"
#include "usbi.h"

static char usb_path[PATH_MAX + 1];

int usb_os_open(usb_dev_handle *dev)
{
  char filename[PATH_MAX + 1];

  snprintf(filename, sizeof(filename) - 1, "%s/%03d/%03d",
    usb_path, dev->bus->busnum, dev->device->devicenum);

  dev->fd = open(filename, O_RDWR);
  if (dev->fd < 0) {
    dev->fd = open(filename, O_RDONLY);
    if (dev->fd < 0) {
      fprintf(stderr, "usb_os_open: failed to open %s: %s\n", filename,
        strerror(errno));
      return USB_ERROR;
    }
  }

  return USB_OK;
}

int usb_os_close(usb_dev_handle *dev)
{
  if (dev->fd <= 0)
    return USB_OK;

  if (close(dev->fd) == -1)
    perror("usb_os_close: tried closing device file descriptor");

  return USB_OK;
}

int usb_set_configuration(usb_dev_handle *dev, int configuration)
{
  int ret;

  ret = ioctl(dev->fd, IOCTL_USB_SETCONFIG, &configuration);
  if (ret < 0) {
    fprintf(stderr, "usb_set_configuration: could not set config %d: %s\n",
      configuration, strerror(errno));
    return USB_ERROR;
  }

  dev->config = configuration;

  return USB_OK;
}

int usb_claim_interface(usb_dev_handle *dev, int interface)
{
  int ret;

  ret = ioctl(dev->fd, IOCTL_USB_CLAIMINTF, &interface);
  if (ret < 0) {
    fprintf(stderr, "usb_claim_interface: could not claim intf %d: %s\n",
      interface, strerror(errno));
    return USB_ERROR;
  }

  dev->interface = interface;

  return USB_OK;
}

int usb_set_altinterface(usb_dev_handle *dev, int alternate)
{
  int ret;
  struct usb_setinterface setintf;

  if (dev->interface < 0)
    return USB_ERROR;

  setintf.interface = dev->interface;
  setintf.altsetting = alternate;

  ret = ioctl(dev->fd, IOCTL_USB_SETINTF, &setintf);
  if (ret < 0) {
    fprintf(stderr, "usb_set_altinterface: could not set intf %d/%d: %s\n",
      dev->interface, alternate, strerror(errno));
    return USB_ERROR;
  }

  dev->altsetting = alternate;

  return USB_OK;
}

int usb_bulk_write(usb_dev_handle *dev, int ep, char *bytes, int size,
	int timeout)
{
  struct usb_bulktransfer bulk;
  int ret, sent = 0;

  do {
    bulk.ep = ep;
    bulk.len = size - sent;
    bulk.timeout = timeout;
    bulk.data = (unsigned char *)bytes + sent;
	
    ret = ioctl(dev->fd, IOCTL_USB_BULK, &bulk);
    if (ret < 0) {
      perror("usb_bulk_write: ioctl(IOCTL_USB_BULK)");
      return USB_ERROR;
    }
    sent += ret;
  } while (ret > 0 && sent < size);
  
  return USB_OK;
}

int usb_bulk_read(usb_dev_handle *dev, int ep, char *bytes, int size,
	int timeout)
{
  struct usb_bulktransfer bulk;
  int ret, retrieved = 0;
  
  do {
    bulk.ep = ep;
    bulk.len = size - retrieved;
    bulk.timeout = timeout;
    bulk.data = (unsigned char *)bytes + retrieved;

    ret = ioctl(dev->fd, IOCTL_USB_BULK, &bulk);
    if (ret < 0) {
      perror("usb_bulk_read: ioctl(IOCTL_USB_BULK)");
      return USB_ERROR;
    }
    retrieved += ret;
  } while (ret > 0 && retrieved < size);

  return retrieved;
}

int usb_control_msg(usb_dev_handle *dev, int requesttype, int request,
	int value, int index, char *bytes, int size, int timeout)
{
  struct usb_ctrltransfer ctrl;
  int ret;

  ctrl.bRequestType = requesttype;
  ctrl.bRequest = request;
  ctrl.wValue = value;
  ctrl.wIndex = index;
  ctrl.wLength = size;

  ctrl.data = bytes;
  ctrl.timeout = timeout;

  ret = ioctl(dev->fd, IOCTL_USB_CONTROL, &ctrl);

  return (ret < 0) ? USB_ERROR : USB_OK;
}

int usb_find_devices_on_bus(struct usb_bus *bus)
{
  DIR *dir;
  struct dirent *entry;
  char dirpath[PATH_MAX + 1];

  snprintf(dirpath, PATH_MAX, "%s/%03d", usb_path, bus->busnum);

  dir = opendir(dirpath);
  if (!dir) {
    if (usb_debug)
      fprintf(stderr, "usb_find_devices_on_bus: Couldn't opendir %s\n",
		dirpath);
    return USB_ERROR;
  }

  while ((entry = readdir(dir)) != NULL) {
    char filename[PATH_MAX + 1];
    struct usb_device *dev;
    int fd, ret;

    /* Skip anything starting with a . */
    if (entry->d_name[0] == '.')
      continue;

    dev = malloc(sizeof(*dev));
    if (!dev) {
      if (usb_debug)
        fprintf(stderr, "usb_find_devices_on_bus: Couldn't allocate memory " \
		" device\n");

      return USB_ERROR;
    }

    memset((void *)dev, 0, sizeof(*dev));

    dev->bus = bus;

    sprintf(filename, "%s/%s", dirpath, entry->d_name);
    fd = open(filename, O_RDONLY);
    if (fd < 0) {
      if (usb_debug)
        fprintf(stderr, "usb_find_devices_on_bus: Couldn't open %s\n",
		filename);
      continue;
    }

    ret = read(fd, (void *)&dev->descriptor, sizeof(dev->descriptor));

    close(fd);

    if (ret < 0) {
      if (usb_debug)
        fprintf(stderr, "usb_find_devices_on_bus: Couldn't read descriptor\n");

      free(dev);
      continue;
    }

    dev->devicenum = atoi(entry->d_name);

    if (bus->devices) {
      dev->next = bus->devices;
      dev->next->prev = dev;
    }
    bus->devices = dev;

    if (usb_debug >= 2)
       fprintf(stderr, "usb_find_devices_on_bus: Found device %d on bus %d\n",
		dev->devicenum, bus->busnum);
  }

  closedir(dir);

  return 0;
}

int usb_find_busses(void)
{
  DIR *dir;
  struct dirent *entry;

  dir = opendir(usb_path);
  if (!dir) {
    if (usb_debug)
      fprintf(stderr, "usb_find_busses: Couldn't opendir %s\n", usb_path);

    return USB_ERROR;
  }

  while ((entry = readdir(dir)) != NULL) {
    struct usb_bus *bus;
    int ret;

    /* Skip anything starting with a . */
    if (entry->d_name[0] == '.')
      continue;

    ret = strspn(entry->d_name, "0123456789");
    if (ret != strlen(entry->d_name)) {
      if (usb_debug >= 2)
        fprintf(stderr, "usb_find_busses: Skipping non bus directory %s\n",
		entry->d_name);
      continue;
    }

    bus = malloc(sizeof(*bus));
    if (!bus) {
      if (usb_debug)
        fprintf(stderr, "usb_find_busses: Couldn't allocate memory for bus\n");

      return USB_ERROR;
    }

    memset((void *)bus, 0, sizeof(*bus));

    bus->busnum = atoi(entry->d_name);

    if (usb_busses) {
      bus->next = usb_busses;
      bus->next->prev = bus;
    }
    usb_busses = bus;

    if (usb_debug >= 2)
       fprintf(stderr, "usb_find_busses: Found bus %d\n", bus->busnum);
  }

  closedir(dir);

  return 0;
}

static int check_usb_vfs(const unsigned char *dirname)
{
  DIR *dir;
  struct dirent *entry;
  int found = 0;

  dir = opendir(dirname);
  if (!dir)
    return 0;

  while ((entry = readdir(dir)) != NULL) {
    /* Skip anything starting with a . */
    if (entry->d_name[0] == '.')
      continue;

    /* We assume if we find any files that it must be the right place */
    found = 1;
    break;
  }

  closedir(dir);

  return found;
}

void usb_os_init(void)
{
  /* Find the path to the virtual filesystem */
  if (check_usb_vfs("/proc/bus/usb"))
    strcpy(usb_path, "/proc/bus/usb");
  else if (check_usb_vfs("/dev/usb"))
    strcpy(usb_path, "/dev/usb");
  else
    usb_path[0] = 0;	/* No path, no USB support */

  if (usb_debug) {
    if (usb_path[0])
      fprintf(stderr, "usb_os_init: Found USB VFS at %s\n", usb_path);
    else
      fprintf(stderr, "usb_os_init: No USB VFS found, is it mounted?\n");
  }
}

